from __future__ import annotations

import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Generic, Optional, Set, Type, Union

from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.func import entrypoint
from langgraph.types import RetryPolicy
from pydantic import TypeAdapter, ValidationError

from livechain.aio.channel import Chan, ChanClosed
from livechain.aio.utils import cancel_and_wait
from livechain.graph.cron import CronExpr
from livechain.graph.types import (
    CronSignal,
    LangGraphInjectable,
    ReactiveSignal,
    T,
    TEvent,
    TModel,
    TriggerSignal,
    TState,
    WatchedValue,
)

logger = logging.getLogger(__name__)


class Mode:
    @dataclass
    class Interrupt:
        pass

    @dataclass
    class Parallel:
        pass

    @dataclass
    class Queue:
        pass

    @dataclass
    class Debounce:
        delay: float


SignalStrategy = Union[
    Mode.Interrupt,
    Mode.Parallel,
    Mode.Queue,
    Mode.Debounce,
]


class SignalRoutineType(str, Enum):
    SUBSCRIBE = "EventCallback"
    REACTIVE = "ReactiveEffect"
    CRON = "CronEffect"
    WORKFLOW = "Workflow"


@dataclass
class _TaskStatus:
    task: asyncio.Task
    has_started: bool = False


def default_signal_strategy() -> Mode.Parallel:
    return Mode.Parallel()


class BaseSignalRoutine(Generic[TModel], ABC):
    def __init__(
        self,
        schema: Type[TModel],
        routine: Callable[[TModel], Awaitable[None]],
        strategy: Optional[SignalStrategy] = None,
        name: Optional[str] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        self._schema = schema
        self._routine = routine
        self._strategy = strategy or default_signal_strategy()
        self._name = name if name is not None else self._routine.__name__
        self._retry = retry

    @property
    @abstractmethod
    def routine_type(self) -> SignalRoutineType:
        raise NotImplementedError

    @property
    def schema(self) -> Type[TModel]:
        return self._schema

    @property
    def name(self) -> str:
        return self._name

    @property
    def mode(self) -> SignalStrategy:
        return self._strategy

    def create_routine_runnable(
        self,
        injectable: LangGraphInjectable | None = None,
    ) -> Runnable[TModel, Any]:
        from livechain.graph.func import step
        from livechain.graph.func.utils import rename_function

        injectable = injectable or LangGraphInjectable.from_empty()

        @step(name=self._name, retry=self._retry)
        async def routine_step(signal: TModel):
            logger.debug(f"Running routine {self._name} with signal {signal}")
            return await self._routine(signal)

        @entrypoint(
            checkpointer=injectable.checkpointer,
            store=injectable.store,
            config_schema=injectable.config_schema,
        )
        @rename_function(self.routine_type.value)
        async def routine_entrypoint(signal: TModel):
            return await routine_step(signal)

        return routine_entrypoint

    def create_runner(
        self,
        config: RunnableConfig | None = None,
        injectable: LangGraphInjectable | None = None,
    ) -> SignalRoutineRunner[TModel]:
        injectable = injectable or LangGraphInjectable.from_empty()
        routine_runnable = self.create_routine_runnable(injectable)

        runner_cls: Optional[Type[SignalRoutineRunner[TModel]]] = {
            Mode.Interrupt: InterruptableSignalRoutineRunner,
            Mode.Parallel: ParallelSignalRoutineRunner,
            Mode.Queue: FifoSignalRoutineRunner,
            Mode.Debounce: DebounceSignalRoutineRunner,
        }.get(type(self._strategy))

        if runner_cls is None:
            raise ValueError(f"Invalid signal routine strategy: {self._strategy}")

        return runner_cls(
            self._schema,
            routine_runnable,
            self._strategy,
            config,
            self._name,
        )


class WorkflowSignalRoutine(BaseSignalRoutine[TriggerSignal]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.WORKFLOW


class EventSignalRoutine(BaseSignalRoutine[TEvent]):
    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.SUBSCRIBE


class ReactiveSignalRoutine(BaseSignalRoutine[ReactiveSignal[TState]], Generic[TState, T]):
    def __init__(
        self,
        schema: Type[ReactiveSignal[TState]],
        routine: Callable[[ReactiveSignal[TState]], Awaitable[None]],
        state_schema: Type[TState],
        cond: WatchedValue[TState, T],
        name: Optional[str] = None,
        strategy: Optional[SignalStrategy] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        super().__init__(schema, routine, strategy, name, retry)
        self._state_schema = state_schema
        self._cond = cond

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.REACTIVE

    @property
    def cond(self) -> WatchedValue[TState, T]:
        return self._cond

    @property
    def state_schema(self) -> Type[TState]:
        return self._state_schema


class CronSignalRoutine(BaseSignalRoutine[CronSignal]):
    def __init__(
        self,
        schema: Type[CronSignal],
        routine: Callable[[CronSignal], Awaitable[Any]],
        cron_expr: CronExpr,
        strategy: Optional[SignalStrategy] = None,
        name: Optional[str] = None,
        retry: Optional[RetryPolicy] = None,
    ):
        super().__init__(schema, routine, strategy, name, retry)
        self._cron_expr = cron_expr

    @property
    def routine_type(self) -> SignalRoutineType:
        return SignalRoutineType.CRON

    @property
    def cron_expr(self) -> CronExpr:
        return self._cron_expr


class SignalRoutineRunner(Generic[TModel], ABC):
    def __init__(
        self,
        schema: Type[TModel],
        runnable: Runnable[TModel, None],
        strategy: SignalStrategy,
        config: RunnableConfig,
        name: str,
    ):
        self._id = uuid.uuid4()
        self._schema = schema
        self._runnable = runnable
        self._strategy = strategy
        self._config = config
        self._name = name

        self._routine_task: Optional[asyncio.Task] = None
        self._signal_ch = Chan[TModel]()

    @property
    def schema(self) -> Type[TModel]:
        return self._schema

    @property
    def name(self) -> str:
        return self._name

    @property
    def routine_id(self) -> str:
        return str(self._id)

    @property
    def strategy(self) -> SignalStrategy:
        return self._strategy

    async def __call__(self, signal: TModel):
        try:
            adapter = TypeAdapter(self._schema)
            validated_signal = adapter.validate_python(signal)
            await self._signal_ch.send(validated_signal)
        except ValidationError as e:
            logger.error(f"Routine runner {self._name} of id {self.routine_id} received invalid data: {e}")

    @abstractmethod
    async def _handle_signal(self, signal: TModel):
        raise NotImplementedError

    async def _handle_cleanup(self):
        logger.info(f"Stopping routine runner {self._name} of id {self.routine_id}")

    async def _routine_loop(self):
        try:
            while True:
                try:
                    signal = await self._signal_ch.recv()
                    await self._handle_signal(signal)
                except ChanClosed:
                    break
                except Exception as e:
                    logger.error(f"Routine runner {self._name} of id {self.routine_id} received an exception: {e}")
                    raise e
        finally:
            await self._handle_cleanup()

    async def start(self):
        self._routine_task = asyncio.create_task(self._routine_loop())
        await self._routine_task

    async def stop(self):
        self._signal_ch.close()
        if self._routine_task is not None:
            await self._routine_task
            self._routine_task = None


class InterruptableSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._current_task: Optional[asyncio.Task] = None

    async def _handle_signal(self, signal: TModel):
        await try_cancel_asyncio_tasks(self._current_task)
        self._current_task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))

    async def _handle_cleanup(self):
        await super()._handle_cleanup()
        await try_cancel_asyncio_tasks(self._current_task)
        self._current_task = None


class ParallelSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._tasks: Set[asyncio.Task] = set()

    def _on_task_done(self, task: asyncio.Task):
        self._tasks.remove(task)

    async def _handle_signal(self, signal: TModel):
        task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))
        task.add_done_callback(lambda _, t=task: self._on_task_done(t))
        self._tasks.add(task)

    async def _handle_cleanup(self):
        await super()._handle_cleanup()
        await try_cancel_asyncio_tasks(*self._tasks)
        self._tasks.clear()


class FifoSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._current_task: Optional[asyncio.Task] = None

    async def _handle_signal(self, signal: TModel):
        self._current_task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))
        await self._current_task

    async def _handle_cleanup(self):
        await super()._handle_cleanup()
        await try_cancel_asyncio_tasks(self._current_task)
        self._current_task = None


class DebounceSignalRoutineRunner(SignalRoutineRunner[TModel]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._pending_task: Optional[asyncio.Task] = None
        self._running_tasks: Set[asyncio.Task] = set()
        self._delay = self._strategy.delay  # type: ignore

    def _on_task_done(self, task: asyncio.Task):
        self._running_tasks.remove(task)

    async def _try_cancel_pending_task(self):
        await try_cancel_asyncio_tasks(self._pending_task)
        self._pending_task = None

    async def _start_pending_task(self, signal: TModel):
        await asyncio.sleep(self._delay)

        # reset pending task and create the running task
        self._pending_task = None
        task = asyncio.create_task(self._runnable.ainvoke(signal, config=self._config))

        # add the running task to the set, remove when done
        self._running_tasks.add(task)
        task.add_done_callback(lambda _, t=task: self._on_task_done(t))

    async def _handle_signal(self, signal: TModel):
        await self._try_cancel_pending_task()
        self._pending_task = asyncio.create_task(self._start_pending_task(signal))

    async def _handle_cleanup(self):
        await super()._handle_cleanup()
        await try_cancel_asyncio_tasks(self._pending_task, *self._running_tasks)
        self._running_tasks.clear()
        self._pending_task = None


async def try_cancel_asyncio_tasks(*tasks: asyncio.Task | None):
    all_tasks = [t for t in tasks if t is not None]
    await cancel_and_wait(*all_tasks)



import asyncio
import logging
import uuid
from collections.abc import AsyncIterable
from typing import Any, AsyncGenerator

from dotenv import find_dotenv, load_dotenv
from livekit.agents import (
    Agent,
    AgentSession,
    JobContext,
    JobProcess,
    RoomInputOptions,
    RoomOutputOptions,
    WorkerOptions,
    cli,
    metrics,
)
from livekit.agents.llm import FunctionTool
from livekit.agents.llm.chat_context import ChatContext, ChatMessage
from livekit.agents.llm.llm import ChatChunk
from livekit.agents.voice import MetricsCollectedEvent
from livekit.agents.voice.agent import ModelSettings
from livekit.plugins import cartesia, deepgram, openai, silero, turn_detector
from livekit.plugins.openai.utils import to_chat_ctx

from examples.basic_workflow import AgentConfig, CreateReminderEvent, RemindUserEvent, UserChatEvent, create_executor
from examples.utils import convert_oai_message_to_langchain_message
from livechain.graph import WorkflowExecutor
from livechain.graph.types import EventSignal

logger = logging.getLogger("basic-agent")

load_dotenv(find_dotenv())


class MyAgent(Agent):
    def __init__(self, executor: WorkflowExecutor, llm_stream_q: asyncio.Queue[AsyncGenerator[str, None]]) -> None:
        super().__init__(
            instructions="You are a helpful assistant that can answer questions and help with tasks.",
        )
        self.executor = executor
        self.llm_stream_q = llm_stream_q
        self.cache_key = uuid.uuid4()

    async def on_enter(self):
        self.session.say("Hello, my name is Alex. Nice to meet you!")

    async def on_end_of_turn(self, chat_ctx: ChatContext, new_message: ChatMessage, generating_reply: bool) -> None:
        chat_ctx = chat_ctx.copy()
        chat_ctx.items.append(new_message)

        oai_messages = to_chat_ctx(chat_ctx, str(self.cache_key))
        lc_messages = [convert_oai_message_to_langchain_message(oai_message) for oai_message in oai_messages]

        self.executor.publish_event(UserChatEvent(messages=lc_messages))

    async def llm_node(
        self, chat_ctx: ChatContext, tools: list[FunctionTool], model_settings: ModelSettings
    ) -> AsyncIterable[ChatChunk] | None | AsyncIterable[str] | str:
        async for chunk in await self.llm_stream_q.get():
            yield chunk


def prewarm(proc: JobProcess):
    proc.userdata["vad"] = silero.VAD.load()


async def entrypoint(ctx: JobContext):
    await ctx.connect()

    session = AgentSession(
        vad=ctx.proc.userdata["vad"],
        # any combination of STT, LLM, TTS, or realtime API can be used
        llm=openai.LLM(model="gpt-4o-mini"),
        stt=deepgram.STT(model="nova-3"),
        tts=cartesia.TTS(),
        # use LiveKit's turn detection model
        turn_detection=turn_detector.EOUModel(),
    )

    llm_stream_q = asyncio.Queue[AsyncGenerator[str, None]]()
    executor = create_executor()

    @executor.recv("llm_stream")
    async def on_llm_stream(stream: AsyncGenerator[str, None]):
        llm_stream_q.put_nowait(stream)

    @executor.recv("reminder_stream")
    async def on_reminder_stream(stream: AsyncGenerator[str, None]):
        await session.say(stream)

    create_reminder_event = CreateReminderEvent(reset=False)
    create_reset_event = CreateReminderEvent(reset=True)
    cancel_reminder_event = RemindUserEvent(should_remind=False)

    event_name_to_signal = {
        "agent_speech_committed": create_reminder_event,
        "agent_stopped_speaking": create_reminder_event,
        "user_stopped_speaking": create_reset_event,
        "agent_started_speaking": cancel_reminder_event,
        "user_started_speaking": cancel_reminder_event,
    }

    def create_event_handler(event_name: str, signal: EventSignal):
        def on_event(_: Any = None):
            logger.info(f"publishing event {event_name}")
            executor.publish_event(signal)

        return on_event

    for event_name, signal in event_name_to_signal.items():
        session.on(event_name, create_event_handler(event_name, signal))  # type: ignore

    # log metrics as they are emitted, and total usage after session is over
    usage_collector = metrics.UsageCollector()

    @session.on("metrics_collected")
    def _on_metrics_collected(ev: MetricsCollectedEvent):
        metrics.log_metrics(ev.metrics)
        usage_collector.collect(ev.metrics)

    async def log_usage():
        summary = usage_collector.get_summary()
        logger.info(f"Usage: {summary}")

    # shutdown callbacks are triggered when the session is over
    ctx.add_shutdown_callback(log_usage)

    # wait for a participant to join the room
    await ctx.wait_for_participant()

    executor.start(config=AgentConfig(name="Alex"))
    await session.start(
        agent=MyAgent(executor=executor, llm_stream_q=llm_stream_q),
        room=ctx.room,
        room_input_options=RoomInputOptions(
            # uncomment to enable Krisp BVC noise cancellation
            # noise_cancellation=noise_cancellation.BVC(),
        ),
        room_output_options=RoomOutputOptions(transcription_enabled=True),
    )


if __name__ == "__main__":
    cli.run_app(WorkerOptions(entrypoint_fnc=entrypoint, prewarm_fnc=prewarm))



from __future__ import annotations

from typing import Any, Dict, List, cast

from langchain_core.messages import AIMessage, AnyMessage, HumanMessage, SystemMessage, ToolMessage
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)


def convert_oai_message_to_langchain_message(
    message: ChatCompletionMessageParam,
) -> AnyMessage:
    role = message["role"]
    lc_message: AnyMessage | None = None

    if role == "user":
        message = cast(ChatCompletionUserMessageParam, message)
        content = message["content"] if isinstance(message["content"], str) else list(message["content"])
        lc_message = HumanMessage(content=content)  # type: ignore

    elif role == "system":
        message = cast(ChatCompletionSystemMessageParam, message)
        content = (
            message["content"] if isinstance(message["content"], str) else list(message["content"])  # type: ignore
        )
        lc_message = SystemMessage(content=content)  # type: ignore

    elif role == "assistant":
        message = cast(ChatCompletionAssistantMessageParam, message)
        additional_kwargs: Dict = {}

        if "content" not in message or message["content"] is None:
            content = ""
        else:
            content = (
                message["content"] if isinstance(message["content"], str) else list(message["content"])  # type: ignore
            )

        if "function_call" in message:
            additional_kwargs["function_call"] = message["function_call"]

        if "tool_calls" in message:
            additional_kwargs["tool_calls"] = message["tool_calls"]

        if "refusal" in message:
            additional_kwargs["refusal"] = message["refusal"]

        lc_message = AIMessage(content=content, additional_kwargs=additional_kwargs)  # type: ignore

    elif role == "tool":
        message = cast(ChatCompletionToolMessageParam, message)
        content = (
            message["content"] if isinstance(message["content"], str) else list(message["content"])  # type: ignore
        )
        lc_message = ToolMessage(content=content, tool_call_id=message["tool_call_id"])  # type: ignore
    else:
        raise ValueError(f"Unsupported message type: {role}")

    return lc_message


def convert_langchain_messages_to_oai_messages(
    messages: List[AnyMessage],
) -> List[ChatCompletionMessageParam]:
    return [convert_langchain_message_to_oai_message(message) for message in messages]


def convert_langchain_message_to_oai_message(
    message: AnyMessage,
) -> ChatCompletionMessageParam:
    role = message.type
    oai_message: Dict[str, Any]

    if role == "human":
        oai_message = {"role": "user", "content": message.content}

    elif role == "system":
        oai_message = {"role": "system", "content": message.content}

    elif role == "ai":
        oai_message = {"role": "assistant", "content": message.content}

        if "function_call" in message.additional_kwargs:
            oai_message["function_call"] = message.additional_kwargs["function_call"]

        if "tool_calls" in message.additional_kwargs:
            oai_message["tool_calls"] = message.additional_kwargs["tool_calls"]

        if "refusal" in message.additional_kwargs:
            oai_message["refusal"] = message.additional_kwargs["refusal"]

    elif role == "tool":
        message = cast(ToolMessage, message)
        oai_message = {
            "role": "tool",
            "content": message.content,
            "tool_call_id": message.tool_call_id,
        }

    else:
        raise ValueError(f"Unsupported message type: {role}")

    return cast(ChatCompletionMessageParam, oai_message)



import asyncio
import logging
from typing import List

from dotenv import find_dotenv, load_dotenv
from langchain_core.messages import AIMessage, AnyMessage, HumanMessage
from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field

from livechain import root, step, subscribe
from livechain.graph.executor import Workflow
from livechain.graph.func.routine import Mode
from livechain.graph.ops import channel_stream, get_config, get_state, mutate_state, publish_event, trigger_workflow
from livechain.graph.types import EventSignal

load_dotenv(find_dotenv())

logger = logging.getLogger("basic_workflow")


class AgentState(BaseModel):
    messages: List[AnyMessage] = Field(default_factory=list)
    has_reminded: bool = Field(default=False)


class AgentConfig(BaseModel):
    name: str = Field(default="assistant")


class UserChatEvent(EventSignal):
    messages: List[AnyMessage]


class CreateReminderEvent(EventSignal):
    reset: bool


class RemindUserEvent(EventSignal):
    should_remind: bool


@step()
async def call_llm(topic: str):
    state = get_state(AgentState)
    config = get_config(AgentConfig)
    llm = ChatOpenAI(model="gpt-4o-mini")

    system_message = AIMessage(
        content=(
            f"You are {config.name} a voice assistant created by LiveKit. Your interface with users will be voice."
        )
    )

    async with channel_stream(topic) as stream_send:
        async for chunk in llm.astream([system_message, *state.messages]):
            await stream_send(chunk.content)


@subscribe(UserChatEvent)
async def on_user_chat(event: UserChatEvent):
    await mutate_state(messages=event.messages)
    await trigger_workflow()


@subscribe(CreateReminderEvent, strategy=Mode.Queue())
async def on_speech_status_changed(event: CreateReminderEvent):
    if event.reset:
        await mutate_state(has_reminded=False)

    publish_event(RemindUserEvent(should_remind=True))


@subscribe(RemindUserEvent, strategy=Mode.Interrupt())
async def on_remind_user(event: RemindUserEvent):
    if not event.should_remind:
        return

    # debounce reminder
    await asyncio.sleep(10)

    # if reminder has already been sent, do not send again
    if get_state(AgentState).has_reminded:
        return

    user_message = HumanMessage(
        content="(Now user keep been silent for 10 seconds, check if user is still active, you would say:)"
    )
    await mutate_state(messages=[user_message], has_reminded=True)
    await call_llm("reminder_stream")


@root()
async def root_routine():
    logger.info("root routine")
    await call_llm("llm_stream")


def create_executor():
    wf = Workflow.from_routines(root_routine, [on_user_chat, on_remind_user, on_speech_status_changed])
    executor = wf.compile(state_schema=AgentState, config_schema=AgentConfig)
    return executor



import asyncio
import logging
from typing import Any, AsyncGenerator

from dotenv import load_dotenv
from livekit.agents import AutoSubscribe, JobContext, JobProcess, WorkerOptions, cli
from livekit.agents.llm import ChatContext
from livekit.agents.pipeline import VoicePipelineAgent
from livekit.plugins import deepgram, openai, silero, turn_detector

from examples.basic_workflow import AgentConfig, CreateReminderEvent, RemindUserEvent, UserChatEvent, create_executor
from examples.utils import EchoStream, NoopLLM, convert_chat_ctx_to_langchain_messages
from livechain.graph.types import EventSignal

load_dotenv()
logger = logging.getLogger("voice-assistant")


def prewarm(proc: JobProcess):
    proc.userdata["vad"] = silero.VAD.load()


async def entrypoint(ctx: JobContext):
    logger.info(f"connecting to room {ctx.room.name}")
    await ctx.connect(auto_subscribe=AutoSubscribe.AUDIO_ONLY)
    cache_key = ctx.room.name

    # wait for the first participant to connect
    participant = await ctx.wait_for_participant()
    logger.info(f"starting voice assistant for participant {participant.identity}")

    executor = create_executor()
    llm_stream_q = asyncio.Queue()

    @executor.recv("llm_stream")
    async def on_llm_stream(stream: AsyncGenerator[str, None]):
        llm_stream_q.put_nowait(stream)

    @executor.recv("reminder_stream")
    async def on_reminder_stream(stream: AsyncGenerator[str, None]):
        await agent.say(stream)

    async def before_llm_cb(agent: VoicePipelineAgent, chat_ctx: ChatContext):
        lc_messages = convert_chat_ctx_to_langchain_messages(chat_ctx, cache_key)
        ev = UserChatEvent(messages=lc_messages)

        await executor.publish_event(ev)
        stream = await llm_stream_q.get()
        return EchoStream(stream, chat_ctx=chat_ctx, fnc_ctx=None)

    agent = VoicePipelineAgent(
        vad=ctx.proc.userdata["vad"],
        stt=deepgram.STT(),
        llm=NoopLLM(),
        tts=openai.TTS(),
        turn_detector=turn_detector.EOUModel(),
        before_llm_cb=before_llm_cb,
    )

    create_reminder_event = CreateReminderEvent(reset=False)
    create_reset_event = CreateReminderEvent(reset=True)
    cancel_reminder_event = RemindUserEvent(should_remind=False)

    event_name_to_signal = {
        "agent_speech_committed": create_reminder_event,
        "agent_stopped_speaking": create_reminder_event,
        "user_speech_committed": create_reset_event,
        "user_stopped_speaking": create_reset_event,
        "agent_started_speaking": cancel_reminder_event,
        "user_started_speaking": cancel_reminder_event,
        "agent_speech_interrupted": cancel_reminder_event,
    }

    def create_event_handler(event_name: str, signal: EventSignal):
        def on_event(_: Any = None):
            logger.info(f"publishing event {event_name}")
            executor.publish_event(signal)

        return on_event

    for event_name, signal in event_name_to_signal.items():
        agent.on(event_name, create_event_handler(event_name, signal))  # type: ignore

    executor.start(config=AgentConfig(name="Alex"))
    agent.start(ctx.room, participant)

    await agent.say("Hey, how can I help you today?", allow_interruptions=True)


if __name__ == "__main__":
    cli.run_app(
        WorkerOptions(
            entrypoint_fnc=entrypoint,
            prewarm_fnc=prewarm,
        ),
    )
